# Copyright 2007 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import os
import unittest

import gvn.hooks.runner


class HookTests(unittest.TestCase):

  def setUp(self):
    gvn.hooks.runner.SetupLogging('unittest', logfile='/dev/null')

  def testImportHookBogus(self):
    # test non-Python file ('/dev/null')
    # (strips .py off, but it's not there, just leaving 'n')
    self.assertRaises(ImportError,
                      gvn.hooks.runner.ImportHook, '/dev', 'null')

    # test non-Python file ('./testdata/plain_file.py')
    self.assertRaises(SyntaxError,
                      gvn.hooks.runner.ImportHook, 'testdata/hooks',
                                                   'plain_file.py')

  def testImportHookCallable(self):
    argv = ['./testdata/pre-commit', './testdata', '42' ]
    hi = gvn.hooks.info.HookInfo(argv, userdb='ignored')
    hi.hook_dir = './testdata/hooks'  # cheat!
    logger = logging.getLogger(hi.logger_name('test'))

    # test success ('./testdata/py_success.py')
    self.assertEquals(0,
        gvn.hooks.runner.ImportHook(hi.hook_dir,
                                    'py_success.py')(hi, logger))

    # test success ('./testdata/py_failure.py')
    self.assertNotEquals(0,
        gvn.hooks.runner.ImportHook(hi.hook_dir,
                                    'py_failure.py')(hi, logger))

    # test "importable" success of scripts whose names aren't proper python
    # module names ('./testdata/123_success.py')
    self.assertEquals(0,
        gvn.hooks.runner.ImportHook(hi.hook_dir,
                                    '123_success.py')(hi, logger))

  def testListHooks(self):
    argv = ['./testdata/hooks/post-commit', './testdata', '42' ]
    hi = gvn.hooks.info.HookInfo(argv, userdb='ignored')
    # Mock os.listdir to return unsorted list.
    orig = os.listdir
    os.listdir = lambda *a: ['b.py', 'a.py']
    try:
      self.assertEquals(gvn.hooks.runner.ListHooks(hi), ['a.py', 'b.py'])
    finally:
      os.listdir = orig

  def testMain(self):
    argv = ['./testdata/hooks/post-commit', './testdata', '42' ]
    self.assertEquals(0, gvn.hooks.runner.main(argv, print_output=False))
