# Copyright 2007 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import shutil
import unittest

from cStringIO import StringIO

import svn.core, svn.ra

from svn.core import SubversionException

from gvn import changebranch
import gvn.errors
import gvn.platform
from gvn.changebranch import ChangeBranch
from gvn import wc
from testing import testcommon


def MakeChanges(_wc, pool):
  """Change some files.  Return a BranchAndAssert list."""

  paths = []
  _wc.Open(write_lock=True)

  paths.append(('A', 'new'))
  open(paths[-1][1], 'w').write('new file\n')
  testcommon.WCAdd(_wc, paths[-1][1], pool)

  paths.append(('D', 'greek-tree/A/mu'))
  testcommon.WCDelete(_wc, paths[-1][1], pool)

  paths.append(('R', 'greek-tree/A/D/gamma'))
  testcommon.WCDelete(_wc, paths[-1][1], pool)
  open(paths[-1][1], 'w').write('replaced file\n')
  testcommon.WCAdd(_wc, paths[-1][1], pool)

  paths.append(('R', 'greek-tree/A/D/H/psi'))
  testcommon.WCDelete(_wc, paths[-1][1], pool)
  testcommon.WCCopy(_wc, 'greek-tree/A/D/G/tau', paths[-1][1], pool)

  paths.append(('R', 'greek-tree/A/D/H/omega'))
  testcommon.WCDelete(_wc, paths[-1][1], pool)
  testcommon.WCCopy(_wc, 'greek-tree/A/D/G/tau', paths[-1][1], pool)
  open(paths[-1][1], 'a').write('replaced/copied/mod file\n')

  paths.append(('A', 'newdir'))
  os.mkdir(paths[-1][1])
  testcommon.WCAdd(_wc, paths[-1][1], pool)

  paths.append(('A', 'newdir/newfile'))
  open(paths[-1][1], 'w').write('new file in a new dir\n')
  testcommon.WCAdd(_wc, paths[-1][1], pool)

  paths.append(('A', 'newdir/cpfile'))
  testcommon.WCCopy(_wc, 'greek-tree/A/D/H/chi', paths[-1][1], pool)

  paths.append(('A', 'newdir/cpmodfile'))
  testcommon.WCCopy(_wc, 'greek-tree/A/D/H/chi', paths[-1][1], pool)
  open(paths[-1][1], 'a').write('copied and edited file in a new dir\n')

  paths.append(('M', 'greek-tree/iota'))
  svn.wc.prop_set2('fooprop', 'quux', _wc.AbsolutePath(paths[-1][1]),
                   svn.wc.adm_probe_retrieve(_wc._adm_access,
                                             _wc.AbsolutePath(paths[-1][1]),
                                             pool),
                   False,               # skip_checks
                   pool)
  testcommon.sleepf(lambda: file(paths[-1][1], 'a').write('new line\n'))

  _wc.Close()
  paths.sort(key=lambda x: x[1])
  return paths

def MakeChanges2():
  """Edit rho.  Return a BranchAndAssert list."""

  paths = []

  paths.append(('M', 'greek-tree/A/D/G/rho'))
  testcommon.sleepf(lambda: open(paths[-1][1], 'a').write('new line\n'))

  paths.sort(key=lambda x: x[1])
  return paths

def _ChangeList(changes):
  result = []
  for i in changes:
    # Drop /changes/user part.
    path = '/'.join(i.path.split('/')[3:])
    result.append((i.action, path))
  result.sort(key=lambda x: x[1])
  return result

# TODO(epg): Rename gvn.wc.IsModified to IsDirty and move this over there.
def IsModified(status):
  """Return whether the svn_wc_status2_t indicates a modified node."""
  return svn.wc.status_modified in [status.text_status, status.prop_status]

class TestCase(unittest.TestCase):

  def setUp(self):
    self.config = testcommon.GetConfig()
    self.repo = testcommon.NewTmpRepository()
    self.repo_url = testcommon.TmpRepositoryURL()

    self.wcdir = testcommon.NewWorkingCopy()
    os.chdir(self.wcdir)
    self.assert_(os.path.isdir('./.svn'))

    self.pool = svn.core.Pool()
    self.wc = wc.WorkingCopy(self.wcdir, cancel_func=None, notify_func=None,
                             config=self.config, pool=self.pool)
    self.wc.Open()
    testcommon.GetAuthBaton(self.config, self.pool, self.wc.project.repository)

  def BranchAndAssert(self, name, paths, description=None):
    """Create/update a ChangeBranch(name) w/ paths.  Return a ChangeBranch.

    paths is a list of tuples of (action, path) where path is a path
    relative to the project top level and action is what to expect in
    RevisionPath.action for this path in the branch revision.

    """

    head = self.wc.project.repository.GetHead()
    change = ChangeBranch(self.config, self.wc.project, name)

    branch_paths = [x[1] for x in paths]

    # Get paths being removed from changebranch, so we can assert
    # they're really gone.
    orig_paths = set(change.ChangedPathsRelative())
    unbranching = orig_paths.difference(branch_paths)

    self.wc.Open(branch_paths)

    # Get status here rather than leaving it up to change.Branch so we
    # can assert some things later.
    path_status = {}
    def status_cb(target, status):
      path_status[self.wc.RelativePath(target)] = status
    self.wc.Status(branch_paths, recursive=False, get_all=False,
                   no_ignore=False, show_children=False, callback=status_cb)

    # Branch!
    change.Branch(self.wc, path_status, description,
                  pool=self.pool)
    head += 1
    self.head = head

    # basic asserts
    self.assertEquals(change.revision.number, head)
    self.assertEquals(change.change_name,
                      '%s/%s@%d' % (self.wc.project.repository.username,
                                    name, head))
    if description is not None:
      self.assertEquals(change.description, description)
    (revision,) = self.wc.project.repository.GetRevisions(start=head)
    self.assertEquals(revision['svn:log'], change.description)

    new_paths = list(change.ChangedPathsRelative())
    # Assert paths removed are really removed.
    for path in unbranching:
      self.assertFalse(path in new_paths)
      self.assertFalse(path in self.wc.change_state)
    # Assert changebranched paths are in wc state.
    for path in new_paths:
      self.assertEquals(self.wc.change_state[path].change_name, change.name)

    revpaths = dict((x.path.lstrip('/'), x) for x in revision.paths)

    # Path leading to the change branch may or may not be created in
    # this same revision.
    path = self.wc.project.ChangeBranchPath(change.username)
    try:
      del revpaths[path]
    except KeyError:
      pass
    path = self.wc.project.ChangeBranchPath(change.username, change.name)
    try:
      del revpaths[path]
    except KeyError:
      pass

    # The project dir is an A on the first commit to the changebranch,
    # R on subsequent commits.
    cbp = self.wc.project.ChangeBranchPath(change.username, change.name,
                                           'projects')
    self.assertTrue(revpaths[cbp].action in ['A', 'R'])
    source_revision = revpaths[cbp].copyfrom_revision
    del revpaths[cbp]

    for (action, path) in paths:
      cbp = self.wc.project.ChangeBranchPath(change.username, change.name,
                                             '/'.join(['projects', path]))
      # Check that we made the correct action.
      # Append the path to what we check, to make for a clear failure message.
      a = ' '.join([action, cbp])
      b = ' '.join([revpaths[cbp].action, cbp])
      self.assertEquals(a, b)
      # Now check to make sure we Replaced the file on the
      # changebranch correctly (r469 caused us to omit the copyfrom).
      if (source_revision != path_status[path].entry.revision
          and (gvn.wc.IsConflicted(path_status[path])
               or IsModified(path_status[path]))):
        entry = path_status[path].entry
        self.assertEqual(revpaths[cbp].copyfrom_path,
                         entry.url[len(self.repo_url)+1:])
        self.assertEqual(revpaths[cbp].copyfrom_revision, entry.revision)
      # We checked this one, clear it.
      del revpaths[cbp]
    self.assertEquals(len(revpaths), 0)

    return change

  # TODO(epg): Unify with BranchAndAssert.
  def SubmitAndAssert(self, change, paths, description=None):
    # Save it, as .Submit sets it to None.
    change_desc = change.description
    change_name = change.change_name
    change_revision = change.revision

    head = self.wc.project.repository.GetHead()
    submitted = change.Submit(self.wc, description=description, pool=self.pool)
    head += 1
    self.head = head
    self.assertEquals(submitted, head)

    top = 'public/projects'
    (revision,) = self.wc.project.repository.GetRevisions([top],
                                    start=self.wc.project.repository.GetHead(),
                                                          limit=1)
    self.assertEquals(revision.number, head)
    self.assertEquals(revision['gvn:change'], change_name)

    if description is None:
      self.assertEquals(revision['svn:log'], change_desc)
    else:
      self.assertEquals(revision['svn:log'], description)

    revpaths = dict((x.path.lstrip('/'), x) for x in revision.paths)

    # The project dir is a D.
    cbp = self.wc.project.ChangeBranchPath(change.username, change.name)
    self.assertEquals(revpaths[cbp].action, 'D')
    del revpaths[cbp]

    for (action, path) in paths:
      ppath = '/'.join([top, path])
      # Join these together for good failure message.
      a = ' '.join([action, ppath])
      b = ' '.join([revpaths[ppath].action, ppath])
      self.assertEquals(a, b)
      del revpaths[ppath]
    self.assertEquals(revpaths, {})

    # Assert paths are removed from wc state.
    for (action, path) in paths:
      self.assertFalse(path in self.wc.change_state)

    # Assert changebranch marked submitted.
    self.assertEqual(change_revision['gvn:submitted'], str(revision.number))

    return submitted

class TestChangeBranch(TestCase):
  def testBranchesAndSubmits(self):
    """intermixed Branch/Submit, including mixed-revision working copies"""

    self.wc.Close()
    paths = MakeChanges(self.wc, self.pool)
    change1 = self.BranchAndAssert('Foo', paths, 'Foo desc')

    paths2 = MakeChanges2()
    change2 = self.BranchAndAssert('change2', paths2, 'change2 desc')

    overlap = list(paths2)
    overlap.append(paths[0])
    self.assertRaises(gvn.errors.ChangeBranch,
                      self.BranchAndAssert, 'overlap', paths2)

    # Re-branch Foo.
    self.wc.Close()
    self.BranchAndAssert('Foo', paths)
    # Re-branch Foo with new description.
    change1 = self.BranchAndAssert('Foo', paths, 'updated Foo desc')

    self.SubmitAndAssert(change2, paths2)

    # wc top is at 2
    # rho is at 7
    # Edit rho again changebranch it (copying projects@2 then
    # replacing rho with rho@7 + delta), and submit it.
    rho = """This is the file 'rho'.
new line
new line
"""
    rho = rho.splitlines()
    paths2 = MakeChanges2()
    self.assertEquals(open('greek-tree/A/D/G/rho').read().splitlines(), rho)
    change2 = self.BranchAndAssert('change2', paths2, 'change2 desc')
    submitted = self.SubmitAndAssert(change2, paths2)

    self.assertEquals(open('greek-tree/A/D/G/rho').read().splitlines(), rho)
    self.assertEquals(open('greek-tree/A/D/G/.svn/text-base/rho.svn-base').read().splitlines(), rho)
    fp = StringIO()
    svn.ra.get_file(self.wc.project.repository.ra,
                    'public/projects/greek-tree/A/D/G/rho', submitted, fp,
                    self.pool)
    self.assertEquals(fp.getvalue().splitlines(), rho)

    self.SubmitAndAssert(change1, paths)

    # Make sure svn.wc.process_committed3 really worked; commit again.
    paths2 = MakeChanges2()
    change2 = self.BranchAndAssert('change22', paths2, 'again')
    self.SubmitAndAssert(change2, paths2)

  def testOutOfDateParent(self):
    self.wc.Close()

    # First we'll make a new directory.
    paths1 = [('A', 'greek-tree/newdir')]
    # Then we'll make a new file in the new directory and edit iota.
    paths2 = [('A', 'greek-tree/newdir/newfile'),
              ('M', 'greek-tree/iota')]

    # Make and add the new directory.
    testcommon.sleepf(os.mkdir, paths1[0][1])
    self.wc.Open(write_lock=True)
    testcommon.WCAdd(self.wc, paths1[0][1], self.pool)
    self.wc.Close()
    # Branch and submit it.
    change1 = self.BranchAndAssert('make-newdir', paths1, 'change1 desc')
    self.SubmitAndAssert(change1, paths1)

    # Add and edit files.
    open(paths2[0][1], 'w').write('new file in a new dir\n')
    self.wc.Open(write_lock=True)
    testcommon.WCAdd(self.wc, paths2[0][1], self.pool)
    self.wc.Close()
    fp = open(paths2[1][1], 'a')
    fp.write('new line\n')
    testcommon.sleepf(fp.close)
    # Try to branch it; must fail.
    try:
      self.BranchAndAssert('edit-and-add', paths2, 'change2 desc')
    except gvn.errors.OutOfDateParent, e:
      self.wc.Close()
      self.assertEqual(e.branch_path, 'greek-tree')
      testcommon.RunSvn('up', e.branch_path)
      self.BranchAndAssert('edit-and-add', paths2, 'change2 desc')
    else:
      self.fail('gvn.errors.OutOfDateParent not raised')

  def testTextBaseCorruption(self):
    # gvn change writes out a text base!
    self.wc.Close()
    paths = MakeChanges2()
    change = self.BranchAndAssert('foo', paths, None)

    # Leaving behind this tmp file was the ultimate cause of the bug.
    if os.path.exists('greek-tree/A/D/G/.svn/tmp/text-base/rho.svn-base'):
      self.fail('tmp text base left behind')

    # Remove the first line of rho.
    rho = '\n'.join(open('greek-tree/A/D/G/rho').read().splitlines()[1:])
    open('greek-tree/A/D/G/rho', 'w').write(rho)

    # Now submit, which would install the wrong tmp text-base we used
    # to leave behind.
    self.SubmitAndAssert(change, paths)

    # This has always been true; assert just in case.
    self.assertEquals(open('greek-tree/A/D/G/rho').read(), rho)

    # This is what was broken; because of my earlier self.fail,
    # though, we wouldn't actually get here.
    text_base = open('greek-tree/A/D/G/.svn/text-base/rho.svn-base').read()
    self.assertEquals(text_base, rho)

  def testRecursiveAdd(self):
    """branching and snapshotting added trees"""

    # Make a copy.
    self.wc.Close()
    self.wc.Open(write_lock=True)
    testcommon.sleepf(testcommon.WCCopy, self.wc, 'greek-tree', 'foo', self.pool)
    self.wc.Close()
    paths = [('A', 'foo')]
    change1 = self.BranchAndAssert('change1', paths)
    self.assertEquals([(x.action, x.path) for x in change1.changed_paths],
                      [('A', '/changes/unittest/change1/projects/foo')])
    # Snapshot and make sure it's still good.
    change1.Branch(self.wc, pool=self.pool)
    self.assertEquals([(x.action, x.path) for x in change1.changed_paths],
                      [('A', '/changes/unittest/change1/projects/foo')])
    self.SubmitAndAssert(change1, paths)

    # TODO(epg): We need to update for now; see TODO file.
    testcommon.RunSvn('update')

    # Make a copy into that, and add a new tree under it.
    paths = [('A', 'foo/foo'), ('A', 'foo/foo/RecursiveAdd'),
             ('A', 'foo/foo/RecursiveAdd/dir1'),
             ('A', 'foo/foo/RecursiveAdd/file1'),
             ('A', 'foo/foo/RecursiveAdd/dir1/file2')]
    self.wc.Open(write_lock=True)
    testcommon.WCCopy(self.wc, 'greek-tree', paths[0][1], self.pool)
    os.mkdir(paths[1][1])
    os.mkdir(paths[2][1])
    open(paths[3][1], 'w').write('hi\n')
    testcommon.sleepf(open(paths[4][1], 'w').write, 'bye\n')
    for i in paths[1:]:
      testcommon.WCAdd(self.wc, i[1], self.pool)
    self.wc.Close()
    change2 = self.BranchAndAssert('change2', paths)
    self.assertEquals(sorted((x.action, x.path) for x in change2.changed_paths),
                      sorted((action,
                              '/changes/unittest/change2/projects/' + path)
                             for (action, path) in paths))
    # Snapshot and make sure it's still good.
    change2.Branch(self.wc, pool=self.pool)
    self.assertEquals(sorted((x.action, x.path) for x in change2.changed_paths),
                      sorted((action,
                              '/changes/unittest/change2/projects/' + path)
                             for (action, path) in paths))

    self.SubmitAndAssert(change2, paths)

    # TODO(epg): We need to update for now; see TODO file.
    testcommon.RunSvn('update')

    # Modify RecursiveAdd/dir1/file2 again.
    paths = [('M', 'foo/foo/RecursiveAdd/dir1/file2')]
    self.wc.Open(write_lock=True)
    testcommon.sleepf(open(paths[0][1], 'a').write, 'new line\n')
    self.wc.Close()
    change3 = self.BranchAndAssert('change3', paths)
    self.assertEquals(sorted((x.action, x.path) for x in change3.changed_paths),
                      sorted((action,
                              '/changes/unittest/change3/projects/' + path)
                             for (action, path) in paths))
    # Snapshot and make sure it's still good.
    change3.Branch(self.wc, pool=self.pool)
    self.assertEquals(sorted((x.action, x.path) for x in change3.changed_paths),
                      sorted((action,
                              '/changes/unittest/change3/projects/' + path)
                             for (action, path) in paths))
    self.SubmitAndAssert(change3, paths)

  def testRecursiveRemove(self):
    """branching and snapshotting removed trees"""

    # Remove a directory and its children.
    self.wc.Close()
    self.wc.Open(write_lock=True)
    paths = [('D', 'greek-tree/A/D')]
    testcommon.WCDelete(self.wc, paths[0][1], self.pool)
    self.wc.Close()
    change1 = self.BranchAndAssert('change1', paths)
    self.assertEquals(sorted((x.action, x.path) for x in change1.changed_paths),
                      sorted((action,
                              '/changes/unittest/change1/projects/' + path)
                             for (action, path) in paths))
    change1.Branch(self.wc, pool=self.pool)
    self.assertEquals(sorted((x.action, x.path) for x in change1.changed_paths),
                      sorted((action,
                              '/changes/unittest/change1/projects/' + path)
                             for (action, path) in paths))
    self.SubmitAndAssert(change1, paths)

    # Note that you have to update the parent first before you can add
    # a new node of different kind.
    testcommon.RunSvn('update')

    # Add a file at path.
    paths = [('A', 'greek-tree/A/D')]
    testcommon.sleepf(open(paths[0][1], 'w').write, 'hi\n')
    self.wc.Open(write_lock=True)
    testcommon.WCAdd(self.wc, paths[0][1], self.pool)
    self.wc.Close()
    change2 = self.BranchAndAssert('change2', paths)
    self.assertEquals(sorted((x.action, x.path) for x in change2.changed_paths),
                      sorted((action,
                              '/changes/unittest/change2/projects/' + path)
                             for (action, path) in paths))
    change2.Branch(self.wc, pool=self.pool)
    self.assertEquals(sorted((x.action, x.path) for x in change2.changed_paths),
                      sorted((action,
                              '/changes/unittest/change2/projects/' + path)
                             for (action, path) in paths))
    self.SubmitAndAssert(change2, paths)

  def testOutOfDate(self):
    """try to submit out of date file"""

    # Change and branch rho.
    paths = MakeChanges2()
    change = self.BranchAndAssert('ood', paths)

    # Now edit rho in the repository.
    os.mkdir('tmp')
    testcommon.RunSvn('co', '--username', 'basil', '--password', 'rayjandom',
           self.repo_url + '/public/projects/greek-tree/A/D/G', 'tmp/G')
    open('tmp/G/rho', 'a').write('yo\n')
    # Windows doesn't build argv quite right, and thus complains that the log
    # message is a directory if it's left empty.
    testcommon.RunSvn('ci', '--username', 'basil', '--password', 'rayjandom',
                      '-m', 'log message', 'tmp/G')

    # Branch ood again; this works fine, and results in a new
    # identical branch.
    change = self.BranchAndAssert('ood', paths)

    # Now try to submit; this should fail with out of date error.
    try:
      change.Submit(self.wc, pool=self.pool)
    except SubversionException, e:
      self.assertTrue(e.apr_err in gvn.errors.SVN_OUT_OF_DATE,
                    msg='Submit failed with %s rather than out-of-date' % (e,))
    except Exception, e:
      self.fail('Submit failed with %s rather than out-of-date' % (e,))
    else:
      self.fail('Submit succeeded rather than fail out-of-date')

  def testBranchFromRoot(self):
    self.wc.Close()
    self.wcdir += '2'
    testcommon.RunSvn('co', self.repo_url, self.wcdir)
    os.chdir(self.wcdir)
    gamma = 'public/projects/greek-tree/A/D/gamma'
    fp = open(gamma, 'a')
    fp.write('new line\n')
    testcommon.sleepf(fp.close)

    self.wc = wc.WorkingCopy(self.wcdir, cancel_func=None, notify_func=None,
                             config=self.config, pool=self.pool)
    testcommon.GetAuthBaton(self.config, self.pool, self.wc.project.repository)
    head = self.wc.project.repository.GetHead()
    change = ChangeBranch(self.config, self.wc.project, 'from-root')
    self.assertRaises(gvn.errors.BranchFromRoot,
                      change.Branch, self.wc, [gamma], pool=self.pool)

  def testDelete(self):
    self.wc.Close()
    paths = MakeChanges(self.wc, self.pool)
    change1 = self.BranchAndAssert('Foo', paths, 'Foo desc')
    change1.Delete(self.wc)

    (revision,) = self.wc.project.repository.GetRevisions(
                                    start=self.wc.project.repository.GetHead())
    self.assertEquals([(x.action, x.path) for x in revision.paths],
                      [('D', '/changes/unittest/Foo')])

  def testApprove(self):
    paths = MakeChanges2()
    change = self.BranchAndAssert('ood', paths)

    self.wc = wc.WorkingCopy(self.wcdir, cancel_func=None, notify_func=None,
                             config=self.config, pool=self.pool)
    self.wc.Open()
    self.wc.project.repository._username = 'basil'
    testcommon.GetAuthBaton(self.config, self.pool, self.wc.project.repository)
    change = ChangeBranch(self.config, self.wc.project, 'ood', 'unittest')

    self.assertRaises(KeyError, self.wc.project.repository.GetRevProp,
                      change.revision.number, 'gvn:approve:basil')
    change.Approve(self.pool)
    revprop = self.wc.project.repository.GetRevProp(change.revision.number,
                                                    'gvn:approve:basil')
    self.assertEquals(revprop, '')

  def testWCNeedsSnapshot(self):
    paths = MakeChanges2()
    change = testcommon.sleepf(self.BranchAndAssert, 'foo', paths, None)
    self.wc.Open(write_lock=False)
    def cb(target, status):
      self.assertFalse(self.wc.NeedsSnapshot(self.wc.RelativePath(target),
                                             status, self.pool),
                       'needs snapshot')
    self.wc.Status([path for (action, path) in paths], callback=cb)

    open(paths[0][1], 'a').write('a new line\n')
    def cb(target, status):
      self.assertTrue(self.wc.NeedsSnapshot(self.wc.RelativePath(target),
                                            status, self.pool),
                      'does not need snapshot')
    self.wc.Status([path for (action, path) in paths], callback=cb)

    change = testcommon.sleepf(self.BranchAndAssert, 'foo', paths, None)
    self.wc.Open(write_lock=False)
    def cb(target, status):
      self.assertFalse(self.wc.NeedsSnapshot(self.wc.RelativePath(target),
                                             status, self.pool),
                       'needs snapshot')
    self.wc.Status([path for (action, path) in paths], callback=cb)

    self.wc.Close()
    self.wc.Open(write_lock=True)
    self.SubmitAndAssert(change, paths)
    self.wc.Open(write_lock=False)
    def cb(target, status):
      self.assertRaises(gvn.errors.NotChangeBranched,
                        self.wc.NeedsSnapshot, self.wc.RelativePath(target),
                        status, self.pool)
    self.wc.Status([path for (action, path) in paths], callback=cb)

  def testListChangeBranches(self):
    self.wc.Close()

    paths = MakeChanges(self.wc, self.pool)
    self.BranchAndAssert('change1', paths)
    change1_name = 'unittest/change1@' + str(self.head)
    self.assertEquals(sorted(x for (x,y) in
                             changebranch.ListChangeBranches(self.wc.project,
                                                             user='unittest',
                                                             pool=self.pool)),
                      [change1_name])
    self.assertEquals(sorted(x for (x,y) in
                             changebranch.ListChangeBranches(self.wc.project,
                                                             pool=self.pool)),
                      [change1_name])

    paths2 = MakeChanges2()
    self.BranchAndAssert('change2', paths2)
    change2_name = 'unittest/change2@' + str(self.head)
    self.assertEquals(sorted(x for (x,y) in
                             changebranch.ListChangeBranches(self.wc.project,
                                                             user='unittest',
                                                             pool=self.pool)),
                      [change1_name, change2_name])
    self.assertEquals(sorted(x for (x,y) in
                             changebranch.ListChangeBranches(self.wc.project,
                                                             pool=self.pool)),
                      [change1_name, change2_name])

    # Make a change as basil.
    self.wc = wc.WorkingCopy(self.wcdir, cancel_func=None, notify_func=None,
                             config=self.config, pool=self.pool)
    self.wc.Open()
    self.wc.project.repository._username = 'basil'
    testcommon.GetAuthBaton(self.config, self.pool, self.wc.project.repository)
    self.BranchAndAssert('change1', paths)
    basil_change1_name = 'basil/change1@' + str(self.head)

    self.assertEquals(sorted(x for (x,y) in
                             changebranch.ListChangeBranches(self.wc.project,
                                                             user='basil',
                                                             pool=self.pool)),
                      [basil_change1_name])
    self.assertEquals(sorted(x for (x,y) in
                             changebranch.ListChangeBranches(self.wc.project,
                                                             pool=self.pool)),
                      [basil_change1_name, change1_name, change2_name])
    self.BranchAndAssert('change2', paths2)
    basil_change2_name = 'basil/change2@' + str(self.head)
    self.assertEquals(sorted(x for (x,y) in
                             changebranch.ListChangeBranches(self.wc.project,
                                                             user='basil',
                                                             pool=self.pool)),
                      [basil_change1_name, basil_change2_name])
    self.assertEquals(sorted(x for (x,y) in
                             changebranch.ListChangeBranches(self.wc.project,
                                                             pool=self.pool)),
                      [basil_change1_name, basil_change2_name,
                       change1_name, change2_name])

    # Test for a username for which a changebranch path does not exist.
    self.assertRaises(gvn.errors.NoUserChangeBranch, sorted,
                      changebranch.ListChangeBranches(self.wc.project,
                                                      user='manuel', pool=self.pool))

    # Hack the change-branch-base to be not "changes" for this part.
    self.wc.project._project_config['change-branch-base'] = 'fawlty'
    self.assertRaises(gvn.errors.NoChangeBranchBase, sorted,
                      changebranch.ListChangeBranches(self.wc.project, pool=self.pool))
    self.assertRaises(gvn.errors.NoUserChangeBranch, sorted,
                      changebranch.ListChangeBranches(self.wc.project,
                                                   user='unittest', pool=self.pool))

  def testUniqueChangeName(self):
    # TODO(epg): uh yeah...
    pass

  def testRandomBranchName(self):
    self.assertEquals(4, len(changebranch.RandomBranchName()))
    self.assertEquals(8, len(changebranch.RandomBranchName(8)))

  # TODO(epg): Diff functions need much more testing...

  def testDiffSnapshotToWC(self):
    paths2 = MakeChanges2()
    change2 = self.BranchAndAssert('change2', paths2, 'change2 desc')

    # Test --from-change.
    rho = 'greek-tree/A/D/G/rho'
    fp = open(rho, 'w')
    testcommon.sleepf(fp.write, 'replacement\n')
    fp.close()
    expect = (
      'Index: greek-tree/A/D/G/rho\n'
      '===================================================================\n'
      '--- greek-tree/A/D/G/rho\t(^/changes/unittest/change2/projects/greek-tree/A/D/G/rho@%d)\n'
      '+++ greek-tree/A/D/G/rho\t(working copy)\n'
      '@@ -1,2 +1 @@\n'
      "-This is the file 'rho'.\n"
      '-new line\n'
      '+replacement\n'
      % (change2.revision.number,))
    got = []
    class mock_file(object):
      def write(fp, line):
        got.append(line)
    encoding = 'utf8'
    callbacks = gvn.changebranch.SvnDiffCallbacks(options=[],
                                                  out_file=gvn.platform.StrippedCRFile(mock_file()),
                                                  encoding=encoding,
                                                  pool=self.pool)
    gvn.changebranch.DiffSnapshotToWC(change2, self.wc, callbacks, encoding,
                                      self.pool)
    self.assertEqual(''.join(got), expect)

    # Test --from-change when not at the wc top.
    os.chdir('greek-tree/A')
    got = []
    gvn.changebranch.DiffSnapshotToWC(change2, self.wc, callbacks, encoding,
                                      self.pool)
    self.assertEqual(''.join(got), expect)
    os.chdir('../..')

  def testDiffChangesAfterCopy(self):
    # Test a case that used to break gvn: copy a tree and then add
    # and/or remove paths under the copied-to tree.  In the past, this
    # has variously exploded or shown nothing, depending on the phase
    # of the moon.  See testStatusShowChildren in test_wc for more,
    # and TODO for still-broken submit.
    self.wc.Close()
    self.wc.Open(write_lock=True)
    testcommon.WCCopy(self.wc, 'greek-tree/A/D', 'acopy', self.pool)
    testcommon.WCCopy(self.wc, 'acopy/gamma', 'acopy/gamma-hey', self.pool)
    open('acopy/gamma-hey', 'a').write('add a line\n')
    testcommon.WCDelete(self.wc, 'acopy/gamma', self.pool)
    self.wc.Close()

    self.wc.Open(write_lock=False)
    paths = dict((self.wc.RelativePath(target), status) for (target, status) in
                 self.wc.Status(['acopy'], recursive=True, show_children=False,
                                get_all=False))
    change = ChangeBranch(self.config, self.wc.project, 'cp-then-rm')
    change.Branch(self.wc, paths, None, self.pool)

    expect = (
      'Index: acopy/gamma (deleted)\n'
      '===================================================================\n'
      'Index: acopy/gamma-hey\n'
      '===================================================================\n'
      '--- acopy/gamma-hey\t(added)\n'
      '+++ acopy/gamma-hey\t(^/changes/unittest/cp-then-rm/projects/acopy/gamma-hey@%d)\n'
      '@@ -1 +1,2 @@\n'
      " This is the file 'gamma'.\n"
      '+add a line\n'
      % (change.revision.number,))
    got = []
    class mock_file(object):
      def write(fp, line):
        got.append(line)
    encoding = 'utf8'
    callbacks = gvn.changebranch.SvnDiffCallbacks(options=[],
                                                  out_file=gvn.platform.StrippedCRFile(mock_file()),
                                                  encoding=encoding,
                                                  pool=self.pool)
    gvn.changebranch.DiffBaseToSnapshot(change, callbacks, encoding, self.pool)
    self.assertEqual(''.join(got), expect)

  def testDiff(self):
    self.wc.Close()
    # Build CreateAndAssert paths structure.
    paths = [
      ('D', 'greek-tree/A/B'),
      ('R', 'greek-tree/A/mu'),
      ('M', 'greek-tree/A/D/H/chi'),
      ('A', 'greek-tree/B.new'),
      ('A', 'greek-tree/B.new/D.new'),
      ('M', 'greek-tree/B.new/D.new/gamma'),
      ('A', 'greek-tree/B.new/iota'),
      ('M', 'greek-tree/B.new/E/alpha'),
      ('M', 'greek-tree/B.new/F'),
      ('R', 'greek-tree/iota'),
    ]
    # Rename A/B.
    testcommon.RunSvn('move', 'greek-tree/A/B', 'greek-tree/B.new')
    # Replace A/mu with a copy of gamma
    testcommon.RunSvn('rm', 'greek-tree/A/mu')
    testcommon.RunSvn('copy', 'greek-tree/A/D/gamma', 'greek-tree/A/mu')
    # Copy A/D.
    testcommon.RunSvn('copy', 'greek-tree/A/D', 'greek-tree/B.new/D.new')
    # Edit A.new/D.new/gamma.
    fp = open('greek-tree/B.new/D.new/gamma', 'a')
    fp.write('hi\n')
    fp.close()
    # Copy iota.
    testcommon.RunSvn('copy', 'greek-tree/iota', 'greek-tree/B.new')
    # Edit B.new/E/alpha.
    fp = open('greek-tree/B.new/E/alpha', 'a')
    fp.write('hi\n')
    fp.close()
    # Edit B.new/F.
    testcommon.RunSvn('ps', 'testprop', 'testval', 'greek-tree/B.new/F')
    # Replace iota with new file
    testcommon.RunSvn('rm', 'greek-tree/iota')
    fp = open('greek-tree/iota', 'w')
    fp.write('This is the replaced iota.\n')
    fp.close()
    testcommon.RunSvn('add', 'greek-tree/iota')
    # Edit A/D/H/chi, but also make it binary.
    testcommon.RunSvn('ps', 'svn:mime-type', 'application/octet-stream',
                      'greek-tree/A/D/H/chi')
    fp = open('greek-tree/A/D/H/chi', 'a')
    fp.write('Append pseudo-binary data.\n')
    fp.close()

    # Create changebranch.
    cb = self.BranchAndAssert('Diff', paths, 'Diff desc')

    expect = (
      'Index: greek-tree/A/D/H/chi\n'
      '===================================================================\n'
      'Cannot display: file marked as a binary type.\n'
      '\nProperty changes on: greek-tree/A/D/H/chi\n'
      '___________________________________________________________________\n'
      'Added: svn:mime-type\n'
      '   + application/octet-stream\n'
      'Index: greek-tree/A/mu\n'
      '===================================================================\n'
      '\nProperty changes on: greek-tree/A/mu\n'
      '___________________________________________________________________\n'
      'Added: svn:mergeinfo\n'
      '   + \n'
      '\nProperty changes on: greek-tree/B.new\n'
      '___________________________________________________________________\n'
      'Added: svn:mergeinfo\n'
      '   + \n'
      '\nProperty changes on: greek-tree/B.new/D.new\n'
      '___________________________________________________________________\n'
      'Added: svn:mergeinfo\n'
      '   + \n'
      'Index: greek-tree/B.new/D.new/gamma\n'
      '===================================================================\n'
      '--- greek-tree/A/D/gamma\t(^/public/projects/greek-tree/A/D/gamma@%d)\n'
      '+++ greek-tree/B.new/D.new/gamma\t(^/changes/unittest/Diff/projects/greek-tree/B.new/D.new/gamma@%d)\n'
      '@@ -1 +1,2 @@\n'
      " This is the file 'gamma'.\n"
      '+hi\n'
      'Index: greek-tree/B.new/E/alpha\n'
      '===================================================================\n'
      '--- greek-tree/A/B/E/alpha\t(^/public/projects/greek-tree/A/B/E/alpha@%d)\n'
      '+++ greek-tree/B.new/E/alpha\t(^/changes/unittest/Diff/projects/greek-tree/B.new/E/alpha@%d)\n'
      '@@ -1 +1,2 @@\n'
      " This is the file 'alpha'.\n"
      '+hi\n'
      '\nProperty changes on: greek-tree/B.new/F\n'
      '___________________________________________________________________\n'
      'Added: testprop\n'
      '   + testval\n'
      'Index: greek-tree/B.new/iota\n'
      '===================================================================\n'
      '\nProperty changes on: greek-tree/B.new/iota\n'
      '___________________________________________________________________\n'
      'Added: svn:mergeinfo\n'
      '   + \n'
      # TODO(next-release): When we drop old-style changebranches, this
      # will be a diff as if it were Modified; that's good, just what svn
      # does; don't be alarmed, just update this.
      'Index: greek-tree/iota\n'
      '===================================================================\n'
      '--- greek-tree/iota\t(added)\n'
      '+++ greek-tree/iota\t(^/changes/unittest/Diff/projects/greek-tree/iota@5)\n'
      '@@ -0,0 +1 @@\n'
      '+This is the replaced iota.\n'
      % (cb.revision.number - 1, cb.revision.number,
         cb.revision.number - 1, cb.revision.number))
    got = []
    class mock_file(object):
      def write(fp, line):
        got.append(line)
    encoding = 'utf8'
    callbacks = gvn.changebranch.SvnDiffCallbacks(options=[],
                                                  out_file=gvn.platform.StrippedCRFile(mock_file()),
                                                  encoding=encoding,
                                                  pool=self.pool)
    gvn.changebranch.DiffBaseToSnapshot(cb, callbacks, encoding, self.pool)
    self.assertEqual(''.join(got), expect)

    # TODO(next-release):
    # - Edit just one of those files, snapshot, make sure everything
    #   still looks right.
    # - Edit one of the directories, snapshot, make sure everything
    #   still looks right.
    # - Submit the change, make sure it looks right.
